function [nameu,fu,tidecon,xout]=t_tide(xin,varargin);
% T_TIDE harmonic analysis of a time series
%      [nameu,fu,tidecon,xout]=T_TIDE(XIN) computes the tidal analysis 
%      of the (possibly complex) time series XIN.
%
%      Further inputs are optional and are specified as property/values,
%          'interval'         sampling interval (hours), default 1. 
%          
%      These next are required if nodal corrections are to be computed,
%      otherwise not necessary.
%          'start time'       [year,month,day,hour,min,sec]
%                              - min,sec are optional OR 
%                             decimal day (matlab DATENUM scalar)
%          'latitude'         degrees (+north)
%
%      Where to send the output.
%          'output'          where to send printed output:
%                            'none'    (no printed output)
%                            'screen'  (to screen) - default
%                            FILENAME   (to a file)
%
%      Correction factor for prefiltering
%          'prefilt'         FS,CORR
%                            If the time series has been passed through
%                            a pre-filter of some kind (say, to reduce the
%                            low-frequency variability) then the analyzed
%                            constituents will have to be corrected for
%                            this. The correction transfer function (1/filter
%                            transfer function) has (possibly complex) magnitude 
%                            CORR at frequency FS (cph). Corrections of <>50% are 
%                            not applied; it is assumed these refer to tidal 
%                            constituents that were intentionally filtered out, 
%                            e.g. the fortnightly components).
%
%       Adjustment for long-term behavior ("secular" behavior)
%          'secular'         'mean'   - assume constant offset (default)
%                            'linear' - get linear trend.
%                     
%       Inference of constituents.
%          'inference'       NAME,REFERENCE,AMPRAT,PHASE_OFFSET
%                            where NAME is an array of the names
%                            of constituents to be inferred, REFERENCE
%                            is an array of the names of references,
%                            and AMPRAT and PHASE_OFFSET are the
%                            amplitude factor and phase offset (in degrees)
%                            from the references. NAME and REFERENCE are Nx4
%                            (max 4 characters in name), and AMPRAT and
%                            PHASE_OFFSET are Nx1 (for scalar time series)
%                            and Nx2 for vector time series (column 1 is
%                            for + frequencies and column 2 for - frequencies).
%
%        Shallow water constituents
%          'shallow'         NAME
%                            a matrix whose rows contain the names of 
%                            shallow-water constituents to look for.
%
%        Resolution criterions for least-squares fit (there is
%          no reason to want to change this)       
%          'rayleigh'         Rayleigh criteria (default 1)
%  
% 
%          It is possible to call tide without using property
%          names in which case the assumed calling sequence
%          is T_TIDE(XIN,INTERVAL,START_TIME,LATITUDE,RAYLEIGH).
%
%
%  OUTPUT: nameu=list of constituents used
%          fu=frequency of tidal constituents (cycles/hr)
%          tidecon=[fmaj,emaj,fmin,emin,finc,einc,pha,epha] for vector xin
%                 =[fmaj,emaj,pha,epha] for scalar (real) xin
%            fmaj,fmin - constituent major and minor axes (same units as xin)       
%            emaj,emin - 95% confidence intervals for fmaj,fmin
%            finc - ellipse orientations (degrees)
%            einc - 95% confidence intervals for finc
%            pha - constituent phases (degrees relative to Greenwich)
%            epha - 95% confidence intervals for pha
%          xout=tidal prediction
%
% Notes: Although missing data can be handled with NaN, it is wise not
% too have too many of them. If your time series has a lot of missing
% data at the beginning and/or end then truncate the input time series.
% (The Rayleigh criterion is applied to frequency intervals calculated as
% the inverse of the input series length).

%
% R. Pawlowicz 8/11/99 - completely rewritten from the transliterated-to-matlab
%                        IOS/Foreman fortran code by B. Beardsley, from whom the
%                        spectral approach to errors also came.

% Parse inputs

ray=1;
dt=1;
fid=1;
stime=[];
lat=[];
corr_fs=[0 1e6];
corr_fac=[1  1];
secular='mean';
inf.iname=[];
inf.irefname=[];
shallownames=[];

k=1;
while length(varargin)>0,
  if isstr(varargin{1}),
    switch lower(varargin{1}(1:3)),
      case 'int',
        dt=varargin{2};
      case 'sta',
        stime=varargin{2};
	if length(stime)>1, 
	  stime=[stime(:)' zeros(1,6-length(stime))]; 
	  stime=datenum(stime(1),stime(2),stime(3),stime(4),stime(4),stime(6));
	end;
      case 'lat',
         lat=varargin{2};
      case 'out',
         filen=varargin{2};
	 switch filen,
	   case 'none',
	     fid=-1;
	   case 'screen',
	     fid=1;
	   otherwise
	     [sout,mesg]=fopen(filen,'w');
	     if fid==-1, error(msg); end;
	  end;
      case 'ray',
         ray=varargin{2};
      case 'pre',
         corr_fs=varargin{2};
	 corr_fac=varargin{3};
         varargin(1)=[];
      case 'sec',
         secular=varargin{2};
      case 'inf',
         inf.iname=varargin{2};
	 inf.irefname=varargin{3};
	 inf.amprat=varargin{4};
	 inf.ph=varargin{5};
	 varargin(1:3)=[];
      case 'sha',
         shallownames=varargin{2};
      otherwise,
         error(['Can''t understand property:' varargin{k}]);
    end;
    varargin([1 2])=[]; 
  else  
    switch k,
      case 1,
        dt=varargin{1};
      case 2,
        stime=varargin{1};
      case 3,
        lat=varargin{1};
      case 4,
        ray=varargin{1};
      otherwise
        error('Too many input parameters');
     end;
     varargin(1)=[];
  end;
  k=k+1;
end;
 
[inn,inm]=size(xin);
xin=xin(:); % makes xin a column vector
nobs=length(xin);

nobsu=nobs-rem(nobs-1,2);% makes series odd to give a center point

% Get the frequencies to use in the harmonic analysis

[nameu,fu,ju,namei,fi,jinf,jref]=constituents(ray/(dt*nobsu),shallownames,inf.iname,inf.irefname);
mu=length(fu);

% Find the good data points (here I assume that in a complex time series
% if u is bad so is v)

gd=find(finite(xin(1:nobsu)));
ngood=length(gd);

fprintf('   Points used: %d of %d\n',ngood,nobs)

t=dt*([1:nobs]'-ceil(nobsu/2));  % Time vector for entire time series


% Now solve for the secular trend plus the analysis. Instead of solving for + and -
% frequencies using exp(i*f*t) I use sines and cosines to keep tc real.
% If the input series is real than this will automatically use real-only
% computation. However, for the analysis it's handy to get the + and
% - frequencies and so that's what we do afterwards.

if secular(1:3)=='lin',
  tc=[ones(length(t),1) t*(2/dt/nobsu) cos((2*pi)*t*fu') sin((2*pi)*t*fu') ];

  coef=tc(gd,:)\xin(gd);

  z0=coef(1);dz0=coef(2);
  ap=(coef(3:mu+2)-i*coef(mu+3:end))/2;  % a+ amplitudes
  am=(coef(3:mu+2)+i*coef(mu+3:end))/2;  % a- amplitudes
else
  tc=[ones(length(t),1) cos((2*pi)*t*fu') sin((2*pi)*t*fu') ];
  
  coef=tc(gd,:)\xin(gd);

  z0=coef(1);dz0=0;
  ap=(coef(2:mu+1)-i*coef(mu+2:end))/2;  % a+ amplitudes
  am=(coef(2:mu+1)+i*coef(mu+2:end))/2;  % a- amplitudes
end;

% Check variance explained (but do this withe the original fit)

xout=tc*coef;  % This is the time series synthesized from the analysis
xres=xin-xout; % and the residuals!


if isreal(xin),
  varx=cov(xin(gd));varxp=cov(xout(gd));varxr=cov(xres(gd));
  fprintf('   percent of var removed by fit: %5.2f %%\n',100*(1-varxr/varx));  
else
  varx=cov(real(xin(gd)));varxp=cov(real(xout(gd)));varxr=cov(real(xres(gd)));
  fprintf('   percent of X var removed by fit: %5.2f %%\n',100*(1-varxr/varx));

  vary=cov(imag(xin(gd)));varyp=cov(imag(xout(gd)));varyr=cov(imag(xres(gd)));
  fprintf('   percent of Y var removed by fit: %5.2f %%\n',100*(1-varyr/vary));
end;

% estimate uncertainities in coefficients

xr=fixgaps(xres); % Fill in "internal" Nan with linearly interpolated values
                  % so we can fft things.
		  
[eap,eam]=errcoef(xr(finite(xr)),fu,dt);


% Correct for prefiltering.

corrfac=interp1(corr_fs,corr_fac,fu);
corrfac(corrfac>1.5 | corrfac <.5 | isnan(corrfac))=1;

ap=ap.*corrfac;
am=am.*conj(corrfac);


% apply nodal corrections and calculate phase relative to Greenwich.
if ~isempty(lat) & ~isempty(stime), 
    
  % Get nodal corrections at midpoint time
  [v,u,f]=t_vuf(stime+floor(nobsu./2)./24.0*dt,[ju;jinf],lat);
  
  vu=(v+u)*360; % total phase correction (degrees)
  nodcor=['nodal corrections applied to amplitude and phase relative to center time'];  
  disp(['   ',nodcor]);
else
  f=ones(length(ju)+length(jinf),1);vu=0*f;
  nodcor=['nodal corrections not applied'];  
  disp(['   ',nodcor]);
end

% Any inference?
ii=find(finite(jref));
if ii,
  snarg=nobsu*pi*(fi(ii)   -fu(jref(ii)) )*dt;
  scarg=sin(snarg)./snarg;
 
  if size(inf.amprat,2)==1,
    pearg=     2*pi*(vu(mu+ii)-vu(jref(ii))+inf.ph(ii))/360;
    pcfac=inf.amprat(ii).*f(mu+ii)./f(jref(ii)).*exp(i*pearg);
    pcorr=1+pcfac.*scarg;
    mcfac=conj(pcfac);
    mcorr=conj(pcorr);
  else
    pearg=     2*pi*(vu(mu+ii)-vu(jref(ii))+inf.ph(ii,1))/360;
    pcfac=inf.amprat(ii,1).*f(mu+ii)./f(jref(ii)).*exp(i*pearg);
    pcorr=1+pcfac.*scarg;
    mearg=    -2*pi*(vu(mu+ii)-vu(jref(ii))+inf.ph(ii,2))/360;
    mcfac=inf.amprat(ii,2).*f(mu+ii)./f(jref(ii)).*exp(i*mearg);
    mcorr=1+mcfac.*scarg;
  end;
    
  ap(jref(ii))=ap(jref(ii))./pcorr;       % Changes to existing constituents
  ap=[ap;ap(jref(ii)).*pcfac];            % Inferred constituents

  am(jref(ii))=am(jref(ii))./mcorr;
  am=[am;am(jref(ii)).*mcfac];

  fu=[fu;fi(ii)];
  nameu=[nameu;namei(ii,:)];
  eam=[eam;eam(jref(ii))];  % Error levels remain the same?
  eap=[eap;eap(jref(ii))];  % (what would make more sense?)
end;

% Angle/magnitude form:

epsp=angle(ap)*180/pi;
epsm=angle(am)*180/pi;
ap=abs(ap);
am=abs(am);


% Estimate error in coefficients
%
% We can think of each "a" as a vector in the complex plane, it's tip "fuzzy" and
% lying within a circle of radius ea. 
%
% To get major and minor axes we add/subtract the absolute values of the a+ and a-
% amplitudes. If ea << |a| then the error in |a| is approximately normal, and we
% can just add variances in the usual way. If ea \approx |a| or < |a| then things are
% more complicated!
emaj=sqrt(eap.^2+eam.^2);
emin=emaj;

% To get the inclination and phase we just add/subtract angles (divided by 2).
% If ea << |a| then the angles are distributed approximately normally and
% we can just add variances again.
epha=sqrt(atan2(eap,ap).^2 + atan2(eam,am).^2)*(180/pi)/2;
einc=epha;



% Apply nodal corrections and compute ellipse parameters
aap=ap./f;
aam=am./f;

fmaj=aap+aam; % major axis
fmin=aap-aam; % minor axis

gp=mod( vu-epsp ,360); % positive Greenwich phase in degrees 
gm=mod( vu+epsm ,360);  % negative Greenwich phase in degrees 

finc=mod( (epsp+epsm)/2 , 180); % ellipse inclination in degrees
                                % (mod 180 to prevent ambiguity, i.e. we
				% always ref. against northern semi-major axis.
pha=mod( gp+finc ,360); % greenwich phase in degrees

snr=(fmaj.^2)./(emaj.^2); % signal to noise ratio


if isreal(xin),
    tidecon=[fmaj,emaj,pha,epha];
else
    tidecon=[fmaj,emaj,fmin,emin,finc,einc,pha,epha];
end;

% Sort results by frequency (needed if anything has been inferred since these
% are stuck at the end of the list by code above)
if any(finite(jref)),
 [fu,I]=sort(fu);
 nameu=nameu(I,:);
 tidecon=tidecon(I,:);
 snr=snr(I);
end;

% output results


if fid>1,
 fprintf(fid,'\n%s\n',['file name: ',filen]);
elseif fid==1,
 fprintf(fid,'-----------------------------------\n');
end

if fid>0,
  fprintf(fid,'date: %s\n',date);
  fprintf(fid,'nobs = %d,  ngood = %d,  record length (days) = %.2f\n',nobs,ngood,length(xin)*dt/24);
  if ~isempty(stime); fprintf(fid,'%s\n',['start time: ',int2str(stime)]); end
  fprintf(fid,'rayleigh criterion = %.1f\n',ray);
  fprintf(fid,'%s\n',nodcor);
%  fprintf(fid,'\n     coefficients from least squares fit of x\n');
%  fprintf(fid,'\n tide    freq        |a+|       err_a+      |a-|       err_a-\n');
%  for k=1:length(fu);
%    if ap(k)>eap(k) | am(k)>eam(k), fprintf('*'); else fprintf(' '); end;
%    fprintf(fid,'%s  %8.5f  %9.4f  %9.4f  %9.4f  %9.4f\n',nameu(k,:),fu(k),ap(k),eap(k),am(k),eam(k));
%  end
  fprintf(fid,'\nx0= %.3g, x trend= %.3g\n',real(z0),real(dz0));
  fprintf(fid,['\nvar(x)= ',num2str(varx),'   var(xp)= ',num2str(varxp),'   var(xres)= ',num2str(varxr) '\n']);
  fprintf(fid,'percent var predicted= %.1f %%\n',100*varxp/varx);

  if isreal(xin)
    fprintf(fid,'\n     tidal amplitude and phase with 95% error estimates\n');
    fprintf(fid,'\ntide   freq       amp     amp_err    pha    pha_err     snr\n');
    for k=1:length(fu);
      if snr(k)>1, fprintf('*'); else fprintf(' '); end;
      fprintf(fid,'%s %8.5f %9.4f %8.3f %8.2f %8.2f %8.2g\n',nameu(k,:),fu(k),tidecon(k,:),snr(k));
    end
  else
    fprintf(fid,'\nx0= %.3g, x trend= %.3g\n',imag(z0),imag(dz0));
    fprintf(fid,['\nvar(y)= ',num2str(vary),'    var(yp)= ',num2str(varyp),'  var(yres)= ',num2str(varyr) '\n']);
    fprintf(fid,'percent var predicted= %.1f %%\n',100*varyp/vary);
    fprintf(fid,'\n%s\n',['ellipse parameters with 95% error estimates']);
    fprintf(fid,'\n%s\n',['tide   freq      major  emaj    minor   emin     inc    einc     pha    epha      snr']);
    for k=1:length(fu);
      if snr(k)>1, fprintf('*'); else fprintf(' '); end;
      fprintf(fid,'%s %8.5f %6.3f %7.3f %7.3f %6.2f %8.2f %6.2f %8.2f %6.2f %6.2g\n',...
	nameu(k,:),fu(k),tidecon(k,:),snr(k));
    end
    fprintf(fid,['\ntotal var= ',num2str(varx+vary),'   pred var= ',num2str(varxp+varyp) '\n']);
    fprintf(fid,'percent total var predicted= %.1f %%\n\n',100*(varxp+varyp)/(varx+vary));
  end

  if fid~=1, st=fclose(fid); end
end;

xout=reshape(xout,inn,inm);

%---------------------------------------------------------------------
function [nameu,fu,ju,namei,fi,jinf,jref]=constituents(minres,shallow,infname,infref);
% [name,freq,kmpr]=constituents(minres,infname)
% loads tidal constituent table (containing 146 constituents).
% and then picks out only the 'resolvable' frequencies (i.e. those
% that are MINRES apart), base on the comparisons in the third column
% of constituents.dat. Only frequencies in the 'standard' set of
% 69 frequencies are actually used.
% Also return the indices of constituents to be inferred.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 1/11/99
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% If we have the mat-file read it in, otherwise create it and read
% it in!

const=t_getconsts;

ju=find(const.df>=minres);

disp(['   number of standard constituents used: ',int2str(length(ju))])

if ~isempty(shallow),
 for k=1:size(shallow,1),
   j1=strmatch(shallow(k,:),const.name);
   if isempty(j1),
     disp(['Can''t recognize name ' shallow(k,:) ' for forced search']);
   else
     if isnan(const.ishallow(j1)),
       disp([shallow(k,:) ' Not a shallow-water constituent']);
     end;
     disp(['   Forced fit to ' shallow(k,:)]);
     ju=[ju;j1];
   end;
 end;
end;
      
nameu=const.name(ju,:);
fu=const.freq(ju);


% check if neighboring chosen constituents violate the rayleigh criteria.
jck=find(diff(fu)<minres);
if (length(jck)>0)
   disp(['  Warning! Following constituent pairs violate Rayleigh criterion']);
   for ick=1:length(jck);
   disp(['     ',nameu(jck(ick),:),'  ',nameu(jck(ick)+1,:)]);
   end;
end

% For inference, add in list of components to be inferred.

fi=[];namei=[];jinf=[];jref=[];
if ~isempty(infname),
  fi=zeros(size(infname,1),1);
  namei=zeros(size(infname,1),4);
  jinf=zeros(size(infname,1),1)+NaN;
  jref=zeros(size(infname,1),1)+NaN;

  for k=1:size(infname,1),
   j1=strmatch(infname(k,:),const.name);
   if isempty(j1),
     disp(['Can''t recognize name' infname(k,:) ' for inference']);
   else
    jinf(k)=j1;
    fi(k)=const.freq(j1);
    namei(k,:)=const.name(j1,:);
    j1=strmatch(infref(k,:),nameu);
    if isempty(j1),
      disp(['Can''t recognize name ' infref(k,:) ' for as a reference for inference']);
    else
      jref(k)=j1;
      fprintf(['   Inference of ' namei(k,:) ' using ' nameu(j1,:) '\n']);
    end;
   end;
  end;    
  jinf(isnan(jref))=NaN;
end;


%-----------------------------------------------------------------
function [eap,eam]=errcoef(xr,fu,dt)
% [ercx,ersx]=errcoef(xr,fu,dt)computes errors in the cos and sin 
% coefficients of the least square fit to data time series x, based 
% on multitaper spectral analysis of the residual series xres. 
%
%  INPUT:  xres=residual time series for x     
%          fu=tidal frequencies used in fit (cycles/hr)
%          dt=sample time (decimal hours)
%          
%  OUTPUT: eap=error in a+ coefficients
%          eam=error in a- coefficients

% mean values of the noise spectrum are computed for the following 
% 8 frequency bands defined by their center frequency and band width:
% M0 +.1 cpd; M1 +-.2cpd; M2 +-.2cpd; M3 +-.2cpd; M4 +-.2cpd; M5 +-.2cpd;
% M6 +-.21cpd; M7 (.26-.29cpd); and M8 (.30-.50cpd). noise variance within
% each band split equally to cos and sin coefficients.  assumes errors in
% coefficients have normal distributions, so that 95% confidence limits
% are given by +- 1.96 std.  also assumes fu(1)>0.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 1/16/99; 1/20/99
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


mu=length(fu);

% Get a good length for the fft (next power of 2)
nx=length(xr);
nfft=2^max(8,ceil(log2(length(xr))));

[Px,fx]=pmtm(xr,4,nfft,1/dt);
Px=real(Px); % Get rid of small imaginary part

df=fx(3)-fx(2);

fbandp=[.00010 .00417;
        .03192 .04859;
        .07218 .08884;
        .11243 .12910;
        .15269 .16936;
        .19295 .20961;
        .23320 .25100;
        .26000 .50000];

if isreal(xr),
  fbandm=fbandp;    % For a real time series + and - power the same
else
  fbandm=1/dt-fliplr(fbandp); % Translate to range Fs/2-Fs
end;

Pavep=zeros(8,1);
Pavem=zeros(8,1);
for k=8:-1:1
   jband=find(fx>=fbandp(k,1) & fx<=fbandp(k,2));
   if any(jband), Pavep(k)=mean(Px(jband)); else Pavep(k)=Pavep(k+1); end;
   jband=find(fx>=fbandm(k,1) & fx<=fbandm(k,2));
   if any(jband), Pavem(k)=mean(Px(jband)); else Pavem(k)=Pavem(k+1); end;
end

% Take spectral density and multiply it by bandwidth of Fourier bin in
% original signal (different than the bandwidth used here
% by a factor of nfft/nx) to get noise variance for the line component
% in that bin.
erp=1.96.*sqrt(Pavep.*df*nfft/nx); % assumes normal distrubution
erm=1.96.*sqrt(Pavem.*df*nfft/nx); % assumes normal distrubution

eap=zeros(mu,1);
eam=zeros(mu,1);

for k1=1:8;
   k=find(fu>=fbandp(k1,1) & fu<=fbandp(k1,2));
   eap(k)=erp(k1);
   eam(k)=erm(k1);
end


%--------------------------------------------------------
function y=fixgaps(x);
% FIXGAPS Linearly interpolates gaps in a time series
%      YOUT=FIXGAPS(YIN) linearly interpolates over NaN
%      in the input time series (may be complex), but ignores
%      trailing and leading NaN.
%

% R. Pawlowicz 6/Nov/99

y=x;

bd=isnan(x);
gd=find(~bd);

bd([1:(min(gd)-1) (max(gd)+1):end])=0;


y(bd)=interp1(gd,x(gd),find(bd)); 








